@@ -11,7 +11,7 @@ from urllib3.exceptions import InsecureRequestWarning
 class Session:
     def __init__(self,logger):
         """
-        Initalizes a Prisma Cloud API Session Manager.
+        Initializes a Prisma Cloud API Session Manager.
 
         Keyword Arguments:
         logger - optional logger, either pylib logger or loguru
@@ -43,6 +43,7 @@ class Session:
                 res, time_completed = self._api_login()
 
                 retries = 0
+                encountered_401 = False
                 while res.status_code in self.retry_statuses and retries < self.retries:
                     if res.status_code in self.retry_delay_statuses:
                         self.logger.warning(f'FAILED {res.status_code} - {time_completed} seconds')
@@ -60,7 +61,16 @@ class Session:
                         self.logger.warning('Increasing wait time.')
 
                     elif res.status_code == self.expired_code:
+                        if encountered_401 == True:
+                            self.logger.error('ERROR - Can not perform API operations')
+                            self.logger.warning('Steps to troubleshoot: ')
+                            self.logger.warning('1) Ensure Access Key has proper permission.')
+                            self.logger.warning('2) Ensure Prisma Cloud license is valid.')
+                            self.logger.warning('3) If using a CWPP Project, ensure project name is valid in credential configurations.')
+                            exit()
+
                         self._expired_login()
+                        encountered_401 = True
                     else:
                         code = "NO_CODE"
                         try:
@@ -75,6 +85,7 @@ class Session:
                         self.logger.info(res.url)
                         self.logger.warning('RESPONSE TEXT:')
                         self.logger.info(res.text)
+
                     
                     res, time_completed = self._api_login()
 
@@ -114,79 +125,90 @@ class Session:
     def _api_refresh_wrapper(self):
         res = self.empty_res
 
-        while res == self.empty_res and self.u_count < self.unknown_error_max:
-            try:
-                res, time_completed = self._api_refresh()
-
-                retries = 0
-                while res.status_code in self.retry_statuses and retries < self.retries:
-                    if res.status_code in self.retry_delay_statuses:
-                        self.logger.warning(f'CODE {res.status_code} - {time_completed} seconds')
-                        self.logger.warning(f'HEADERS: {res.headers}')
-                        #Increase timer when ever encounter to slow script execution.
-                        if self.retry_timer == 0:
-                            self.retry_timer = 1
-                        else:
-                            self.retry_timer = self.retry_timer*2
-                            if self.retry_timer >= self.retry_timer_max:
-                                self.retry_timer = self.retry_timer_max
-
-                        self.logger.warning(f'Waiting {self.retry_timer} seconds')
-                        time.sleep(self.retry_timer)
-                        self.logger.warning('Increasing wait time.')
-
-                    elif res.status_code == self.expired_code:
-                        self.logger.error(f'FAILED {self.expired_code} - {time_completed} seconds')
-                        self.logger.warning('Session expired. Generating new Token')
-                        self._api_login_wrapper()
-                        return
-                    else:
-                        code = "NO_CODE"
-                        try:
-                            code = res.status_code
-                        except:
-                            pass
-                        self.logger.error(f'FAILED {code} - {time_completed} seconds')
-                        self.logger.error('ERROR Refreshing Token.')
-                        self.logger.warning('RESPONSE:')
-                        self.logger.info(res)
-                        self.logger.warning('RESPONSE URL:')
-                        self.logger.info(res.url)
-                        self.logger.warning('RESPONSE TEXT:')
-                        self.logger.info(res.text)
-                    
+        refresh_func = getattr(self, "_api_refresh", None)
+        if not callable(refresh_func):
+            return self._api_login_wrapper()
+        else:
+            while res == self.empty_res and self.u_count < self.unknown_error_max:
+                try:
                     res, time_completed = self._api_refresh()
 
-                    retries +=1
-                
-                if retries == self.retries:
-                    self.logger.error('ERROR. Max retires exceeded on JWT refresh. Exiting...')
+                    retries = 0
+                    while res.status_code in self.retry_statuses and retries < self.retries:
+                        if res.status_code in self.retry_delay_statuses:
+                            self.logger.warning(f'CODE {res.status_code} - {time_completed} seconds')
+                            self.logger.warning(f'HEADERS: {res.headers}')
+                            #Increase timer when ever encounter to slow script execution.
+                            if self.retry_timer == 0:
+                                self.retry_timer = 1
+                            else:
+                                self.retry_timer = self.retry_timer*2
+                                if self.retry_timer >= self.retry_timer_max:
+                                    self.retry_timer = self.retry_timer_max
+
+                            self.logger.warning(f'Waiting {self.retry_timer} seconds')
+                            time.sleep(self.retry_timer)
+                            self.logger.warning('Increasing wait time.')
+
+                        elif res.status_code == self.expired_code:
+                            self.logger.error(f'FAILED {self.expired_code} - {time_completed} seconds')
+                            self.logger.warning('Session expired. Generating new Token')
+                            self._api_login_wrapper()
+                            return
+                        else:
+                            code = "NO_CODE"
+                            try:
+                                code = res.status_code
+                            except:
+                                pass
+                            self.logger.error(f'FAILED {code} - {time_completed} seconds')
+                            self.logger.error('ERROR Refreshing Token.')
+                            self.logger.warning('RESPONSE:')
+                            self.logger.info(res)
+                            self.logger.warning('RESPONSE URL:')
+                            self.logger.info(res.url)
+                            self.logger.warning('RESPONSE TEXT:')
+                            self.logger.info(res.text)
+                        
+                        res, time_completed = self._api_refresh()
+
+                        retries +=1
+                    
+                    if retries == self.retries:
+                        self.logger.error('ERROR. Max retires exceeded on JWT refresh. Exiting...')
+                        exit()
+
+                    #Update token and headers
+                    token = res.json().get('token')
+                    self.token = token
+                    new_headers = self.headers
+                    new_headers[self.auth_key] = self.auth_style + token
+                    self.headers = new_headers
+
+                    try:
+                        self.logger.success(f'SUCCESS - {time_completed} seconds')
+                    except:
+                        self.logger.info(f'SUCCESS - {time_completed} seconds')
+
+                    self.u_count = 1
+
+                    return
+
+                except KeyboardInterrupt:
+                    self.logger.error('Keyboard Interrupt. Exiting...')
                     exit()
-
-                try:
-                    self.logger.success(f'SUCCESS - {time_completed} seconds')
-                except:
-                    self.logger.info(f'SUCCESS - {time_completed} seconds')
-
-                self.u_count = 1
-
-                return
-
-            except KeyboardInterrupt:
-                self.logger.error('Keyboard Interrupt. Exiting...')
-                exit()
-            except Exception as e:
-                self.logger.error(e)
-                self.logger.error(f'UNKNOWN ERROR - API refresh. Retrying... {self.u_count} of {self.unknown_error_max}')
-                self.u_count += 1
-                self.logger.warning('Steps to troubleshoot: ')
-                self.logger.warning('1) Disable any VPNs.')
-                self.logger.warning('2) Ensure API base URL is correct.')
-                time.sleep(1)
+                except Exception as e:
+                    self.logger.error(e)
+                    self.logger.error(f'UNKNOWN ERROR - API refresh. Retrying... {self.u_count} of {self.unknown_error_max}')
+                    self.u_count += 1
+                    self.logger.warning('Steps to troubleshoot: ')
+                    self.logger.warning('1) Disable any VPNs.')
+                    self.logger.warning('2) Ensure API base URL is correct.')
+                    time.sleep(1)
 
 
 #==============================================================================
-    def __api_call_wrapper(self, method: str, url: str, json: dict=None, data: dict=None, params: dict=None, verify=True, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[]):
+    def __api_call_wrapper(self, method: str, url: str, json: dict=None, data: dict=None, params: dict=None, verify=True, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[], custom_log='', custom_error_message=''):
         """
         A wrapper around all API calls that handles token generation, retrying
         requests and API error console output logging.
@@ -209,17 +231,29 @@ class Session:
                 res,time_completed = self.__request_wrapper(method, url, headers=self.headers, json=json, data=data, params=params, verify=verify, acceptCsv=acceptCsv)
                 
                 if res.status_code in self.success_status or res.status_code in status_ignore:
+                    log_message = ''
+                    if custom_log:
+                        log_message = ' - ' + custom_log
+
                     try:
-                        self.logger.success(f'SUCCESS - {time_completed} seconds')
+                        self.logger.success(f'SUCCESS - {time_completed} seconds{log_message}')
                     except:
-                        self.logger.info(f'SUCCESS - {time_completed} seconds')
+                        self.logger.info(f'SUCCESS - {time_completed} seconds{log_message}')
                     return res
 
                 retries = 0
+                encountered_401 = False
                 while res.status_code in self.retry_statuses and retries < self.retries:
                     #If we get a 429 code, sleep for a doubling amount of time.
                     if res.status_code in self.retry_delay_statuses:
-                        self.logger.warning(f'FAILED {res.status_code} - {time_completed} seconds')
+                        error_msg = ''
+                        log_message = ''
+                        if custom_error_message:
+                            error_msg = " - " + custom_error_message
+                        if custom_log:
+                            log_message = ' - ' + custom_log
+
+                        self.logger.warning(f'FAILED {res.status_code} - {time_completed} seconds{log_message}{error_msg}')
                         #Wait for retry timer
                         if self.retry_timer > 0:
                             self.logger.warning(f'Waiting {self.retry_timer} seconds')
@@ -236,9 +270,18 @@ class Session:
                     
                     #If token expires, login again and get new token
                     if res.status_code == self.expired_code:
+                        if encountered_401 == True:
+                            self.logger.error('ERROR - Can not perform API operations')
+                            self.logger.warning('Steps to troubleshoot: ')
+                            self.logger.warning('1) Ensure Access Key has proper permission.')
+                            self.logger.warning('2) Ensure Prisma Cloud license is valid.')
+                            self.logger.warning('3) If using a CWPP Project, ensure project name is valid in credential configurations.')
+                            exit()
+
                         self.logger.error(f'FAILED {self.expired_code} - {time_completed} seconds')
                         self.logger.warning('Session expired. Generating new Token and retrying')
                         self._api_login_wrapper()
+                        encountered_401 = True
 
                     self.logger.warning(f'Retrying request')
                     self.logger.debug(f'{url}')
@@ -313,7 +356,7 @@ class Session:
 
     #==============================================================================
 
-    def request(self, method: str, endpoint_url: str, json: dict=None, data: dict=None, params: dict=None, verify=None, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[]):
+    def request(self, method: str, endpoint_url: str, json: dict=None, data: dict=None, params: dict=None, verify=None, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[], custom_log='', custom_error_message=''):
         '''
         Function for calling the PC API using this session manager. Accepts the
         same arguments as 'requests.request' minus the headers argument as 
@@ -322,6 +365,16 @@ class Session:
         if verify == None:
             verify = self.verify
 
+        #If the CWP Session is a project, auto add project id query string
+        try:
+            if self.project_flag == True:
+                if params:
+                    params.update({"project":self.tenant})
+                else:
+                    params = {"project":self.tenant}
+        except:
+            pass
+
         #Validate method
         method = method.upper()
         if method not in ['POST', 'PUT', 'GET', 'OPTIONS', 'DELETE', 'PATCH']:
@@ -334,7 +387,7 @@ class Session:
         url = f'{self.api_url}{endpoint_url}'
 
         #Call wrapper
-        return self.__api_call_wrapper(method, url, json=json, data=data, params=params, verify=verify, acceptCsv=acceptCsv, redlock_ignore=redlock_ignore, status_ignore=status_ignore)
+        return self.__api_call_wrapper(method, url, json=json, data=data, params=params, verify=verify, acceptCsv=acceptCsv, redlock_ignore=redlock_ignore, status_ignore=status_ignore, custom_log=custom_log, custom_error_message=custom_error_message)
 
     def config_search_request(self, json: dict, verify=None, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[]):
         if verify == None:
